/*
 * main_utils.cpp
 *
 *  Created on: May 9, 2016
 *      Author: vlegrand
 */
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/stat.h>
#include <limits.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <cmath>
#include "main_utils.h"

using namespace std;

/*
 * Use this function to get the amount of RAM in GB on the machine (used or not).
 * Aim is to make sure that ROCK can run on the current machine with the requirements that the user gave.
 */
unsigned long getNodePhysMemory() {
    uint64_t mem = 0;

#if defined(_SC_PHYS_PAGES) && defined(_SC_PAGESIZE)
    mem = sysconf(_SC_PHYS_PAGES);
    mem *= sysconf(_SC_PAGESIZE);
    mem /=  1 * 1024 * 1024 * 1024;
#elif __APPLE__
    size_t len = sizeof(mem);
    int ret=sysctlbyname("hw.memsize", (void*) &mem, &len, NULL,0);
    if (ret==-1) {
        cout<<"Couldn't determine how much RAM is usable on your system."<<errno<<endl;
	mem=0;
    }
    mem /=  1 * 1024 * 1024 * 1024;
#else
    cout<<"ROCK has not been tested on this operating system, please contact author."<<endl;
#endif

    return (unsigned long)mem;
}


/*
 * Given a filename (including full path), determine if parent directory exists.
 * If is doesn't, display error message and exit.
 */
std::string checkDirExists(const string o_file) {
    string parent_dir=".";
    std::size_t o_found = o_file.find_last_of("/");
    if (o_found!=string::npos) {
        parent_dir=o_file.substr(0,o_found);
        // cout<<parent_dir<<endl;
        struct stat info;
        if (stat(parent_dir.c_str(),&info)!=0) {
            cout<<"parent directory for output files: "<<parent_dir<<" doesn't exist."<<endl;
            exit(EXIT_FAILURE);
        }
    }
    return parent_dir;
}

/*
 * logarithm of the gamma function;
 * from here: https://www.cec.uchile.cl/cinetica/pcordero/MC_libros/NumericalRecipesinC.pdf
 */

double gammln(const unsigned long& xx) {
	double x,y,tmp,ser;

	static double cof[6]={76.18009172947146,-86.50532032941677,\
			24.01409824083091,-1.231739572450155,\
			0.001208650973866179,-0.000005395239384953};
	int j;
	y=x=xx;
	tmp=x+5.5;
	tmp -=(x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser+=cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}

/*
 * here x is the smallest value of the counter for a given k-mer in the CMS.
 * m is the size of a CMS array (AC: in fact, we use the biggest prime number as an approximation of the array's size).
 */
double pmf(const float&x,const unsigned long& nb_k_mers,const unsigned int& m) {
	double res;
    double lnn=log(nb_k_mers);
    double lnm=log(m);
    double diff=lnn-lnm;
    double p=x*diff;
    diff=double(nb_k_mers)/m;
    //double truc=exp(diff);
    double tmp=p-diff;
    double g=gammln(x+1);
    tmp -=g;
    res=exp(tmp);
    return res;
}

double ccdf(const int& smallest_kappa,const unsigned long& nb_k_mers,const unsigned int& m) {
	double res;
	double s=0.0;
	for (int i=0;i<=smallest_kappa;i++) {
		s+=pmf(i,nb_k_mers,m);
	}
	res=1-s;
	if (res<0) res=0.00000001; // It happens because pmf if an approximation of the probability distribution.
			                   // So it happens that ccdf is negative but very close to 0... As we need a non null value for the next calculation,
	                           // return a very small constant.
	if (res>=0.999) res=0.999; // Need to do that because we compule ln(ccdf) to determine lamda and we need it to be non null otherwise we'd get a division by 0.
	return res;
}

/*
 * Minimize lambda for a given number of k-mers.
 * We want p<=0.01,
 * so choose smallest lambda so that ccdf(kappa_prime,nb_k_mer,array_size) exp lambda<=0.01.
 * Smallest is either low filter either high filter if there is no low filter.
 */
int getBestLambdaForN(const unsigned long& nb_k_mers,const int& smallest_kappa,const unsigned int& m,const float& max_collision_proba) {
	float lnp=log(max_collision_proba);
	float ccdfr=ccdf(smallest_kappa,nb_k_mers,m);
	float lnccdf=log(ccdfr);
	float tmp=lnp/lnccdf;
	int min_l=ceil(tmp);
    if (min_l==0) min_l=1;
	return min_l;
}




float getCollisionProba(const int& smallest_kappa,const unsigned long& nb_k_mers,const unsigned int& arr_cms_size,const int& lambda) {
	float ccfdr=ccdf(smallest_kappa,nb_k_mers,arr_cms_size);
	float fpp=pow(ccfdr,lambda);
	return fpp;
}


void usage(int status) {
cout<<"ROCK v@@@                    Copyright (C) 2016-2021 Institut Pasteur"<<endl;
cout<<endl;
cout<<"Reducing Over-Covering K-mers within FASTQ file(s)"<<endl;
cout<<endl;
cout<<"USAGE: rock [options] [files]"<<endl;
cout<<endl;
cout<<"OPTIONS:"<<endl;
cout<<" -i <file>  file containing the name(s)  of the input FASTQ file(s) to"<<endl;
cout<<"	           process;  single-end: one file name per line;  paired-end:"<<endl;
cout<<"	           two file names  per line  separated by  a comma;  up to 15"<<endl;
cout<<"	           FASTQ file  names can  be specified;  of note,  input file"<<endl;
cout<<"	           name(s) can also be specified as program argument(s)"<<endl;
cout<<" -o <file>  file containing the  name(s) of the  output FASTQ file(s);"<<endl;
cout<<"	           FASTQ file name(s) should be structured in the same way as"<<endl;
cout<<"	           the file specified in option -i."<<endl;
cout<<" -k <int>   k-mer length (default 25)"<<endl;
cout<<" -c <int>   lower-bound k-mer coverage depth threshold (default: 0)"<<endl;
cout<<" -C <int>   upper-bound k-mer coverage depth threshold (default: 70)"<<endl;
cout<<" -l <int>   number of hashing function(s) (default: 4)"<<endl;
cout<<" -n <int>   expected total number of  distinct k-mers within the input"<<endl;
cout<<"	           read sequences; not compatible with option -l."<<endl;
cout<<" -f <float> maximum expected false positive probability when computing"<<endl;
cout<<"	           the optimal number of hashing functions from the number of"<<endl;
cout<<"	           distinct k-mers specified with option -n (default: 0.05)."<<endl;
cout<<" -q <int>   sets as valid  only k-mers  made  up  of  nucleotides with"<<endl;
cout<<"	           Phred score (+33 offset) above this cutoff (default: 0)"<<endl;
cout<<" -m <int>   minimum number of  valid k-mer(s) to  consider a read; all"<<endl;
cout<<"	           non-considered reads are  written into output file(s) with"<<endl;
cout<<"	           extension undefined.fastq (default: 1)"<<endl;
cout<<" -v         verbose mode"<<endl;
cout<<" -h         prints this message and exit"<<endl;
}

/*
void usage(int status) {
  cout<<"usage: rock [options] [args]"<<endl<<endl;
  cout<<"    -i <file>      .... file name containing the list of the FASTQ files to process."<<endl;
  cout<<"    -o <file>      .... file name containing the list of the output FASTQ file names."<<endl;
  cout<<"    -h             .... Print this message and exit."<<endl;
  cout<<"    -k <int>       .... k-mer size. (default 25)."<<endl;
  cout<<"    -c <int>       .... lower coverage threshold (default: 0)."<<endl;
  cout<<"    -C <int>       .... upper coverage threshold (default: 70; max: 65535)."<<endl;
  cout<<"    -f <float>     .... maximum probability of collision in the count min sketch (default: 0.05)"<<endl;
  cout<<"    -l <int>       .... size of the count min sketch (default: at most 4, depending on the available RAM)"<<endl;
  cout<<"    -n <int>       .... (expected) number of distinct k-mers within the processed reads."<<endl;
  cout<<"    -q <int>       .... minimum quality threshold for nucleotides. Default is 0."<<endl;
  cout<<"    -m <int>       .... minimum of valid k-mer threshold for reads. Default is 1."<<endl;
  cout<<"    -v             .... verbose"<<endl;
  exit(status); }*/

